/*
** bregex.c
**
** Binary REGular EXpressions
**
** Regular expressions with support for binary and incomplete data
**
** Author:		Julian Anastasov <ja@ssi.bg>, January 2001
**
**			Released under the GPL version 2
*/

/* 
** REGEX is one BRANCH (not a list of branches separated by '|' as the
** standards require)
** 
** 
** BRANCH is a list of PIECEs
** 
** 
** PIECE is ATOM followed by '*', '+', '?' or bound:
** 
** ATOM *		0 or more matches of the atom
** ATOM +		1 or more matches of the atom
** ATOM ?		0 or 1 matches of the atom
** ATOM {n}		n matches of the atom
** ATOM {min,}		min or more matches of the atom
** ATOM {min,max}	min to max matches of the atom
** 
** 
** ATOM is:
** 
** .			match any single char
** list of strings separated by '|':
** (text1|text2)	match text1 or text2
** brace expression:
** [chars]		match specified chars
** [^chars]		don't match the specified chars
** single characters:
** char			match specified char
** escape_char char	character following escape character
** 
** 
** Examples (with escape character '!'):
** 
** []a]			match one char if it is ']' or 'a'
** [^]a]		match any char except ']' or 'a'
** [!!]			match '!'
** [a-z]		match letters from 'a' to 'z'
** 
** to match a special character it must be escaped:
** !*			match '*'
** (a!||b!))		match 'a|' or 'b)'
** [a!]]		match one char if it is 'a' or ']'
**
** Bounds:
** a{0,2}		0 to 2 matches for 'a'
** a{5}			5 matches for 'a'
** (alpha){2,}		2 or more matches for 'alpha'
**
**
** Notes about brace expressions ('[chars]'):
** - strict case matching is performed
** - '^' as first char inverts the result
** - ']' used after the first '[' or following the first '^' is not treated
** as special character. It can be escaped too.
** - escaping in '[chars]' is required only for ']' but can be applied for
** any character
**
** Notes about bounds:
** - no escaping is required in '{bounds}'
** - the bounds are internally stored as 'unsigned long'
** - the negative values are converted to unsigned
** - (long) -1 is reserved and can't be used as value for min or max
** - min <= max
** - min >= 0
**
** Notes about texts:
** - escaping is required for the special characters: the escape character,
** '[', '(', '*', '+', '?', '.'
** - in (text1|text2) escaping is required only for the escape character,
** '|' and ')'
**
*/


#include <stdlib.h>
#include <errno.h>

#include "bregex.h"

/*
** Return the size of the regex atom
*/

int bregex_get_atom(char *r, int sz, int escape)
{
char *s;

	if (sz <= 0) return 0;
	if ('.' == *r) return 1;
	if (escape == *r) {
		if (sz > 1) return 2;
		return -1;
	}
	if ('(' == *r) {
		/* (.) */
		if (sz < 3) return -1;
		s = r++;
		if (')' == *r) return -1;
		for (sz--; sz > 0; r++, sz--) {
			if (escape == *r) {
				if (sz <= 2) return -1;
				r ++;
				sz --;
			} else if (')' == *r) {
				return (r-s+1);
			}
		}
		return -1;
	}
	if ('[' == *r) {
		/* [.] */
		if (sz < 3) return -1;
		s = r++;
		sz --;
		if ('^' == r[1]) {
			sz --;
			r ++;
		}
		if (']' == *r) {
			sz --;
			r ++;
		}
		for (; sz > 0; r++, sz--) {
			if (escape == *r) {
				if (sz <= 2) return -1;
				r ++;
				sz --;
			} else if (']' == *r) {
				return (r-s+1);
			}
		}
		return -1;
	}
	return 1;
}

/*
** Match text with ATOM, return the number of matched characters
** Return:
** -1: error
** -2: we need more data
** 0: even with more data we will fail again
** >0: characters matched
*/

int bregex_match_atom(char *s, int sz, char *r, int escape, int igncase)
{
char c;
unsigned char uc;
char *s0;
int sz0, fl;

	if (!sz) return -2;
	if ('.' == *r) {
		return 1;
	}
	if (escape == *r) {
		r ++;
		goto one;
	}

	if ('(' == *r) {
		s0 = s;
		sz0 = sz;
		fl = 0;
		for (;;) {
			r++;
			s = s0;
			sz = sz0;
			for (;; sz--, s++, r++) {
				if ('|' == *r || ')' == *r) {
					goto l_1;
				}
				if (!sz) {
					/* We need more data */
					fl = -2;
					break;
				}
				if (escape == *r) r++;
				if (*r == *s) continue;
				if (!igncase) break;
				c = *r & 0xDF;
				if (((*s & 0xDF) == c) &&
					c >= 'A' && c <= 'Z') continue;
				break;
			}
			while ('|' != *r && ')' != *r) {
				if (escape == *r) r ++;
				r ++;
			}
			if (')' == *r) return fl;
		}
		l_1:
		if (sz == sz0) return -1;
		return sz0 - sz;
	}

	if ('[' == *r) {
		r ++;
		if ('^' == *r) {
			fl = 0;
			r++;
		} else {
			fl = 1;
		}
		if (']' == *r) {
			if (']' == *s) return fl;
			r ++;
			if ('-' == *r) {
				goto range;
			}
		}

		loop:

		if (']' == *r) return !fl;
		if (escape == *r) r++;
		if (*r == *s) return fl;
		r ++;
		if ('-' != *r) goto loop;

		range:
		uc = *((unsigned char *) (r-1));
		if (escape == r[1]) r ++;
		if (*((unsigned char *) s) >= uc &&
			*((unsigned char *) s) <= *((unsigned char *) (r+1))) {
			return fl;
		}
		r += 2;
		goto loop;

	}

one:
	if (*r == *s) return 1;
	if (!igncase) return 0;
	c = *r & 0xDF;
	if (((*s & 0xDF) == c) && c >= 'A' && c <= 'Z') return 1;
	return 0;
}

/*
** Match regular expression
** 
** Args:
** szp, rszp and matches are in/out
** s, szp: string to match
** r, rszp: regular expression
** escape: escape char to use in the regular expressions
** igncase: ignore case when matching characters not in brace expressions
**
** - Always  put 0 in  r[*rszp], sometimes we  read ahead using external
** functions without looking in *rszp, for example with strtoul()
*/

int bregex_match(char *s, int *szp, char *r, int *rszp, unsigned long *matches,
	int escape, int igncase)
{
int rl;
int sz, rsz;
int sz2, rsz2;
int sta, sz0;
int fl;
unsigned long l, min, max, n, nmatches;
char *ep, *ep2;

	sz = *szp;
	rsz = *rszp;
	fl = 0;
	for (;;) {
		rl = bregex_get_atom(r, rsz, escape);
		if (rl < 0) return rl;
		if (!rl) {
			fl = 1;
			break;
		}
		if (rsz > rl) {
			if ('*' == r[rl]) {	/* 0 or more */
				if (!*szp && rsz - rl -1 <= 0) {
					/* Match end of the text */
					fl = 1;
					rsz = 0;
					*matches = 0;
					goto out_match;
				}
				for (;;) {
					sz2 = sz;
					rsz2 = rsz - rl - 1;
					if (rsz2 > 0) {
						sta = bregex_match(s, &sz2,
							r + rl + 1,
							&rsz2, &nmatches,
							escape, igncase);
						if (sta < 0) return sta;
						if (sta) {
							fl = 1;
							sz = sz2;
							rsz = rsz2;
							*matches = 0;
							goto out_match;
						}
					}
					sz0 = bregex_match_atom(s, sz, r,
						escape, igncase);
					if (sz0 <= 0) {
						if (-1 == sz0) return sz0;
						if (!sz0) {
							rsz -= rl + 1;
							r += rl + 1;
							*matches = 0;
							goto loop_regex;
						}
						goto out_match;
					}
					sz -= sz0;
					s += sz0;
					(*matches) ++;
				}
			}
			if ('+' == r[rl]) {	/* 1 or more */
				for (;;) {
					rsz2 = rsz - rl - 1;
					if (*matches && rsz2 > 0) {
						sz2 = sz;
						//if (rsz2 <= 0) goto out_match;
						sta = bregex_match(s, &sz2,
							r + rl + 1, &rsz2,
							&nmatches,
							escape, igncase);
						if (sta < 0) return sta;
						if (sta) {
							fl = 1;
							sz = sz2;
							rsz = rsz2;
							*matches = 0;
							goto out_match;
						}
					}
					sz0 = bregex_match_atom(s, sz, r,
						escape, igncase);
					if (sz0 <= 0) {
						if (-1 == sz0) return sz0;
						if (!sz0 && *matches) {
							rsz -= rl + 1;
							r += rl + 1;
							*matches = 0;
							goto loop_regex;
						}
						goto out_match;
					}
					sz -= sz0;
					s += sz0;
					(*matches) ++;
				}
			}
			if ('?' == r[rl]) {	/* 0 or 1 */
				if (!*szp && rsz - rl -1 <= 0) {
					/* Match end of the text */
					fl = 1;
					rsz = 0;
					*matches = 0;
					goto out_match;
				}
				//we expect *matches to be 0
				sz2 = sz;
				rsz2 = rsz - rl - 1;
				if (rsz2 > 0) {
					sta = bregex_match(s, &sz2,
						r + rl + 1, &rsz2,
						&nmatches, escape, igncase);
					if (sta < 0) return sta;
					if (sta) {
						fl = 1;
						sz = sz2;
						rsz = rsz2;
						*matches = 0;
						goto out_match;
					}
				}
				sz0 = bregex_match_atom(s, sz, r,
					escape, igncase);
				if (sz0 <= 0) {
					if (-1 == sz0) return sz0;
					if (sz0) goto out_match;
				}
				sz -= sz0;
				s += sz0;
				//(*matches) ++;

				rsz -= rl + 1;
				r += rl + 1;
				*matches = 0;
				continue;
			}
			if ('{' == r[rl]) {
				ep = 0;
				errno = 0;
				l = strtoul(r + rl + 1, &ep, 0);
				if (errno || !ep || -1 == l) return -1;
				min = max = l;
				if (',' == *ep) {
					max = -2;
					ep ++;
					if ('}' != *ep) {
						ep2 = 0;
						l = strtoul(ep, &ep2, 0);
						if (errno || !ep2 ||
							-1 == l || l < min)
							return -1;
						ep = ep2;
						max = l;
					}
				}
				if ('}' != *ep++ || rsz < (ep-r)) return -1;
				if (!*szp && rsz - (ep-r) <= 0 &&
					*matches >= min) {
					/* Match end of the text */
					fl = 1;
					rsz = 0;
					*matches = 0;
					goto out_match;
				}
				if (max && min <= max)
				for (n = *matches; n <= max; n++) {
					if (n) {
						sz0 = bregex_match_atom(s, sz,
							r, escape, igncase);
						if (sz0 <= 0) {
							if (-1 == sz0) return sz0;
							if (!sz0 && n >= min) {
								rsz -= ep-r;
								r = ep;
								*matches = 0;
								goto loop_regex;
							}
							goto out_match;
						}
						sz -= sz0;
						s += sz0;
						(*matches) ++;
					}
					rsz2 = rsz - (ep-r);
					if (n >= min && n < max && rsz2 > 0) {
						sz2 = sz;
						//if (rsz2 <= 0) goto out_match;
						sta = bregex_match(s, &sz2, ep,
							&rsz2, &nmatches,
							escape, igncase);
						if (sta < 0) return sta;
						if (sta) {
							fl = 1;
							sz = sz2;
							rsz = rsz2;
							*matches = 0;
							goto out_match;
						}
					}
				}
				rsz -= (ep-r);
				r = ep;
				*matches = 0;
				continue;
			}
		}
		sz0 = bregex_match_atom(s, sz, r, escape, igncase);
		if (sz0 <= 0) {
			if (-1 == sz) return sz0;
			goto out;
		}
		rsz -= rl;
		r += rl;
		sz -= sz0;
		s += sz0;
		*matches = 0;

		loop_regex:
	}

out:
	if (!fl) {
		*rszp = 0;
		return 0;
	}
out_match:
	*szp = sz;
	*rszp = rsz;
	return fl;
}

